@@ -15,9 +15,15 @@ from einops import rearrange
 
 # helper functions
 
+def identity(t):
+    return t
+
 def default(val, def_val):
     return def_val if val is None else val
 
+def rand_true(prob):
+    return random.random() < prob
+
 def singleton(cache_key):
     def inner_fn(fn):
         @wraps(fn)
@@ -207,10 +213,11 @@ class PixelCL(nn.Module):
         projection_hidden_size = 2048,
         augment_fn = None,
         augment_fn2 = None,
+        prob_rand_hflip = 0.25,
         moving_average_decay = 0.99,
         ppm_num_layers = 1,
         ppm_gamma = 2,
-        distance_thres = 0.1, # the paper uses 0.7, but that leads to nearly all positive hits. need clarification on how the coordinates are normalized before distance calculation.
+        distance_thres = 0.7,
         similarity_temperature = 0.3,
         alpha = 1.
     ):
@@ -226,6 +233,7 @@ class PixelCL(nn.Module):
 
         self.augment1 = default(augment_fn, DEFAULT_AUG)
         self.augment2 = default(augment_fn2, self.augment1)
+        self.prob_rand_hflip = prob_rand_hflip
 
         self.online_encoder = NetWrapper(net, projection_size, projection_hidden_size, layer=hidden_layer)
 
@@ -264,7 +272,13 @@ class PixelCL(nn.Module):
         update_moving_average(self.target_ema_updater, self.target_encoder, self.online_encoder)
 
     def forward(self, x):
-        shape, device = x.shape, x.device
+        shape, device, prob_flip = x.shape, x.device, self.prob_rand_hflip
+
+        rand_flip_fn = lambda t: torch.flip(t, dims = (-1,))
+
+        flip_image_one, flip_image_two = rand_true(prob_flip), rand_true(prob_flip)
+        flip_image_one_fn = rand_flip_fn if flip_image_one else identity
+        flip_image_two_fn = rand_flip_fn if flip_image_two else identity
 
         cutout_coordinates_one, _ = cutout_coordinates(x)
         cutout_coordinates_two, _ = cutout_coordinates(x)
@@ -272,13 +286,18 @@ class PixelCL(nn.Module):
         image_one_cutout = cutout_and_resize(x, cutout_coordinates_one)
         image_two_cutout = cutout_and_resize(x, cutout_coordinates_two)
 
+        image_one_cutout = flip_image_one_fn(image_one_cutout)
+        image_two_cutout = flip_image_two_fn(image_two_cutout)
+
         image_one_cutout, image_two_cutout = self.augment1(image_one_cutout), self.augment2(image_two_cutout)
 
         proj_one = self.online_encoder(image_one_cutout)
         proj_two = self.online_encoder(image_two_cutout)
 
         image_h, image_w = shape[2:]
+
         proj_image_shape = proj_one.shape[2:]
+        proj_image_h, proj_image_w = proj_image_shape
 
         coordinates = torch.meshgrid(
             torch.arange(image_h, device = device),
@@ -287,10 +306,15 @@ class PixelCL(nn.Module):
 
         coordinates = torch.stack(coordinates).unsqueeze(0).float()
         coordinates /= math.sqrt(image_h ** 2 + image_w ** 2)
+        coordinates[:, 0] *= proj_image_h
+        coordinates[:, 1] *= proj_image_w
 
         proj_coors_one = cutout_and_resize(coordinates, cutout_coordinates_one, output_size = proj_image_shape)
         proj_coors_two = cutout_and_resize(coordinates, cutout_coordinates_two, output_size = proj_image_shape)
 
+        proj_coors_one = flip_image_one_fn(proj_coors_one)
+        proj_coors_two = flip_image_two_fn(proj_coors_two)
+
         proj_coors_one, proj_coors_two = map(lambda t: rearrange(t, 'b c h w -> (b h w) c'), (proj_coors_one, proj_coors_two))
         pdist = nn.PairwiseDistance(p = 2)
 
@@ -344,5 +368,7 @@ class PixelCL(nn.Module):
 
         # total loss
 
+        positive_pixel_pairs = positive_mask_one_two.sum()
         loss = loss_pix * self.alpha + loss_pixpro
-        return loss
+
+        return loss, positive_pixel_pairs
